from twitter_demographer.components import Component
from sentence_transformers import SentenceTransformer
import umap
import hdbscan

class BaseClustering(Component):
    """
    This BaseClustering is strongly inspired to BERTopic way of creating the clusters

    See: https://github.com/MaartenGr/BERTopic
    """

    def __init__(self, embedding_model):
        """

        :param embedding_model: SentenceBERT embedding model
        """
        super().__init__()
        self.embedding_model = embedding_model

    def outputs(self):
        return ["clusters"]

    def inputs(self):
        return ["text"]

    def infer(self, data):
        texts = data["text"].fillna(" ")

        embeddings = SentenceTransformer(self.embedding_model).encode(texts)

        embeddings = umap.UMAP(n_neighbors=3,
                              min_dist=0.1,
                              metric='cosine').fit_transform(embeddings)

        clusterer = hdbscan.HDBSCAN(min_cluster_size=2)
        cluster_labels = clusterer.fit_predict(embeddings)

        return {"clusters": cluster_labels}




from twitter_demographer.components import Component
from m3inference import M3Inference
import os
from twitter_demographer.support import utils
from multiprocessing import Pool
from appdirs import user_cache_dir
import shutil


class GenderAndAge(Component):
    """
    Provides a wrapper on the m3 dataset to predict age and binary gender in twitter data
    """

    def __init__(self):
        super().__init__()

    def outputs(self):
        return ["screen_name",
                "name",
                "user_id_str",
                "user_id",
                "id",
                "description",
                "profile_image_url"]

    def inputs(self):
        return ["user_id", "description", "profile_image_url"]

    def infer(self, data):
        classifier = InternalGenderAgeFinder()

        results = self.initialize_return_dict()

        data = data[list(results.keys())]

        to_test_data = data.drop_duplicates(subset=["screen_name"])

        to_test_data = to_test_data.to_dict("records")

        cc = classifier.get_attribute_classification(to_test_data, batch_size=4)

        output_classifier = {
            "age": [],
            "gender": [],
            "is_org": []
        }

        for value in data["user_id"]:
            value = str(value)

            output_classifier["age"].append(cc[value]["age"])
            output_classifier["gender"].append(cc[value]["gender"])
            output_classifier["is_org"].append(cc[value]["is_org"])

        return output_classifier


class InternalGenderAgeFinder:

    def __init__(self):
        self.m3 = M3Inference()
        self.m3_noimg = M3Inference(use_full_model=False)

    def get_attribute_classification(self, users, batch_size):
        cache_dir = user_cache_dir("twitter_demographer", "images")

        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)

        imaged_users = []
        without_image_users = []
        images = []

        for user in users:

            if ('user_id_str' in user) and (not ('user_id' in user)):
                user['user_id'] = str(user['user_id_str'])
            try:
                img_file_resize = "{}/{}_224x224.{}".format(cache_dir, user['user_id'],
                                                            utils.get_extension(user['profile_image_url']))
            except:
                img_file_resize = ""

            images.append((user['profile_image_url'], img_file_resize))

        with Pool(4) as p:
            outputs = p.starmap(utils.download_resize_img, images)

        for user, (is_retrieved, a, exist) in zip(users, outputs):

            if ('user_id_str' in user) and (not ('user_id' in user)):
                user['user_id'] = str(user['user_id_str'])

            try:
                img_file_resize = "{}/{}_224x224.{}".format(cache_dir, user['user_id'],
                                                            utils.get_extension(user['profile_image_url']))
            except:
                img_file_resize = ""

            user['img_path'] = img_file_resize
            user['description'] = user['description']
            user['lang'] = 'un'

            if is_retrieved and a:
                imaged_users.append(user)
            else:
                without_image_users.append(user)

        pred_imaged = self.m3.infer(imaged_users, output_format='json', batch_size=batch_size)
        pred_without = self.m3_noimg.infer(without_image_users, output_format='json', batch_size=batch_size)
        fused = {**pred_imaged, **pred_without}

        shutil.rmtree(cache_dir)
        output_classifier = {}
        for key, value in fused.items():
            gender = 'male' if value['gender']["male"] > value['gender']["female"] else 'female'
            is_org = 'non-org' if value["org"]['non-org'] > value["org"]['is-org'] else 'is-org'
            age_dict = {k: v for (k, v) in value["age"].items()}
            age = max(age_dict, key=age_dict.get)

            output_classifier[str(key)] = {"age": age,
                                           "gender": gender,
                                           "is_org": is_org}

        return output_classifier



from tqdm import tqdm
import geocoder
from twitter_demographer.components import Component
import logging



class GeoNamesDecoder(Component):
    """
    Wrappers on the geocoder API to disambiguate users' locations
    """

    def __init__(self, key):
        super().__init__()
        self.key = key

    def outputs(self):
        return ["geo_location_country", "geo_location_address"]

    def inputs(self):
        return ["location"]

    def infer(self, data):
        logger = logging.StreamHandler()
        logger.setLevel(logging.ERROR)
        geo = self.initialize_return_dict()
        pbar = tqdm(total=len(data))

        for val in data["location"]:
            pbar.update(1)
            if val is None:
                geo["geo_location_country"].append(None)
                geo["geo_location_address"].append(None)
            else:
                g = geocoder.geonames(val, key=self.key)
                geo["geo_location_country"].append(g.country)
                geo["geo_location_address"].append(g.address)
        pbar.close()
        return geo



import hashlib
from tqdm import tqdm

class Demographer:
    """
    Main entry point of twitter-demographer.
    """

    def __init__(self, cache_steps=False):
        self.components = []
        self.cache_steps = cache_steps
        self.local_inputs = []

    def add_component(self, component):
        if len(self.components) == 0:
            self.components.append(component)
            self.local_inputs.extend(component.outputs() + component.inputs())  # first run we add both
        else:
            if set(component.inputs()) <= set(self.local_inputs):
                self.components.append(component)
            else:
                raise Exception("Some of the fields are not supported by this composition")

    def infer(self, data):
        pbar = tqdm(total=len(self.components), position=0)
        pbar.set_description("Running Demographer")
        for component in self.components:

            update = component.infer(data)
            for key, value in update.items():
                data[key] = value

            if self.cache_steps:
                data.to_parquet("local.csv", engine="pyarrow")
            del component
            pbar.update(1)

            # add hash(y: str) because hash(x: int) = x
            data["screen_name"] = data["screen_name"].apply(
                lambda x: hashlib.sha3_256(str(hash(x) + hash("hash")).encode()).hexdigest())
        pbar.close()
        drop_keys = ['tweet_ids', 'name', 'user_id_str',
                     'user_id', 'id', 'profile_image_url', 'description']

        for key in drop_keys:
            del data[key]

        data = data.sample(frac=1)

        return data



import numpy as np
import pandas as pd
from contextualized_topic_models.models.ctm import ZeroShotTM
from contextualized_topic_models.utils.data_preparation import TopicModelDataPreparation
from contextualized_topic_models.utils.preprocessing import WhiteSpacePreprocessingStopwords

class TopicModeler:
    """
    A simple topic modeler based on Contextualized Topic Models. This topic model is multilingual by default

    see: https://aclanthology.org/2021.acl-short.96/
    """

    def __init__(self, topic_num, embedding_model="paraphrase-multilingual-mpnet-base-v2"):
        self.topic_num = topic_num
        self.embedding_model = embedding_model


    def infer(self, data):
        df = pd.DataFrame({"text": data["text"].values.tolist()})
        df = df[~df["text"].isna()]

        documents = df["text"].values.tolist()

        sp = WhiteSpacePreprocessingStopwords(documents)
        preprocessed_documents, unpreprocessed_corpus, vocab = sp.preprocess()

        tp = TopicModelDataPreparation(self.embedding_model)

        training_dataset = tp.fit(text_for_contextual=unpreprocessed_corpus, text_for_bow=preprocessed_documents)

        ctm = ZeroShotTM(bow_size=len(tp.vocab), contextual_size=768, n_components=self.topic_num, num_epochs=2)
        ctm.fit(training_dataset)
        topics_predictions = np.argmax(ctm.get_thetas(training_dataset, n_samples=15), axis=1).tolist()

        returned = [None] * len(data)

        for index, value in zip(df.index.values.tolist(), topics_predictions):
            returned[index] = value

        return {"topic": returned}



import abc

import pandas as pd
from tqdm import tqdm
import tweepy
from tweepy.errors import BadRequest
import transformers
from collections import defaultdict
from abc import ABC

transformers.logging.set_verbosity(transformers.logging.ERROR)

class Component(ABC):

    def __init__(self):
        self.output_parameters = self.outputs()

    @abc.abstractmethod
    def outputs(self):
        pass

    @abc.abstractmethod
    def inputs(self):
        pass

    @abc.abstractmethod
    def infer(self, *args):
        pass

    def initialize_return_dict(self):
        """
        A simple function that creates a dictionary with keys = parameters and values empty lists
        :return:
        """
        results = {}

        for param in self.output_parameters:
            results[param] = []
        return results


class Rehydrate(Component):

    def __init__(self, bearer_token):
        """
        :param bearer_token: token from the twitter developer page
        """
        super().__init__()
        self.api = tweepy.Client(bearer_token, wait_on_rate_limit=True)

    def inputs(self):
        return ["tweet_ids"]

    def outputs(self):
        return ['screen_name', 'name', 'location', 'user_id_str',
                'user_id', 'id', 'profile_image_url', 'description',
                'created_at', 'text']

    def infer(self, data):
        results = self.initialize_return_dict()

        pbar = tqdm(total=len(data))
        for tweet_id in data["tweet_ids"]:
            try:
                requested = self.api.get_tweet(tweet_id,
                                               tweet_fields=[
                                                   'id', 'text', 'author_id', 'created_at', 'conversation_id',
                                                   'entities',
                                                   'public_metrics', 'referenced_tweets'
                                               ],
                                               expansions=[
                                                   'author_id', 'referenced_tweets.id',
                                                   'referenced_tweets.id.author_id',
                                               ],
                                               place_fields=['full_name', 'id'],
                                               user_fields=[
                                                   'id', 'name', 'username', 'profile_image_url', 'description',
                                                   'location'
                                               ])
            except BadRequest as e:
                pbar.update(1)
                for key in self.outputs():
                    results[key].append(None)
                continue

            pbar.update(1)
            pbar.set_description("Running Hydrate")

            tweet = requested[0]
            if (len(requested.errors)) > 0:
                # if there are errors, we have a None
                pbar.update(1)
                for key in self.outputs():
                    results[key].append(None)
                continue

            user_includes = requested[1]["users"][0]

            results["text"].append(tweet.text)
            results["description"].append(user_includes.description)
            results["created_at"].append(tweet.created_at)
            results["name"].append(user_includes.name)
            results["screen_name"].append(user_includes.username)
            results["user_id"].append(user_includes.id)
            results["id"].append(user_includes.id)
            results["user_id_str"].append(str(user_includes.id))
            results["profile_image_url"].append(user_includes.profile_image_url.replace("_normal", ""))
            results["location"].append(user_includes.location)

        pbar.close()
        return results

def not_null(param):
    """
    k = pd.DataFrame({"text" : ["a", None, "b"]})

    @not_null("text")
    def method(data):
        return {"new_label" : [1 for _ in data.iterrows()]}

    print(class_.method(k)) # index at 1 should be None

    {'new_label': [1, None, 1]}

    """
    def Inner(func):
        def wrapper(*args, **kwargs):

            data = args[1]
            to_function = data[~(data[param].isna())]
            args = (args[0],) + (to_function,) + args[2:]
            resulted = func(*args, **kwargs)

            new_dict = dict()

            for keys in resulted:
                new_dict[keys] = [None] * len(data)
                for index, value in zip(to_function.index.values.tolist(), resulted[keys]):
                    new_dict[keys][index] = value

            return new_dict
        return wrapper
    return Inner




import urllib.request
from io import BytesIO
from PIL import Image

def get_extension(img_path):
    dotpos = img_path.rfind(".")
    extension = img_path[dotpos + 1:]
    if extension.lower() == "gif":
        return "png"
    return extension

def download_resize_img(url, img_out_path, img_out_path_fullsize=None):
    # url=url.replace("_200x200","_400x400")
    try:
        img_data = urllib.request.urlopen(url)
        img_data = img_data.read()

        if img_out_path_fullsize != None:
            with open(img_out_path_fullsize, "wb") as fh:
                fh.write(img_data)
    except urllib.error.HTTPError as err:

        #logger.warn(url)
        #logger.warn("Error fetching profile image from Twitter. HTTP error code was {}.".format(err.code))
        #os.mknod(img_out_path)
        #raise err
        return False, False, 0
#     print('here')
    except:
        return False, False, 0

    return True, resize_img(BytesIO(img_data), img_out_path, force=True), 1

def resize_img(img_path, img_out_path, filter=Image.BILINEAR, force=False):
    #img = Image.open(img_path).convert("RGB")

    try:
        img = Image.open(img_path).convert("RGB")
        if img.size[0] + img.size[1] < 400 and not force:
            print(f'{img_path} is too small. Skip.')
            return
        img = img.resize((224, 224), filter)
        img.save(img_out_path)
        return True
    except Exception as e:
        print(f'Error when resizing {img_path}\nThe error message is {e} for {img_out_path}\n')
        #logger.warning(f'Error when resizing {img_path}\nThe error message is {e} for {img_out_path}\n')
        return False




def tt(tokenizer):
    def tokenize_function(examples):
        # Remove empty lines
        examples["texts"] = [
            line
            for line in examples["texts"]
            if len(line) > 0 and not line.isspace()
        ]

        return tokenizer(
            examples["texts"],
            padding=True, truncation=True)
    return tokenize_function

def prepare_dataset(dataset, tokenizer):
    dataset = dataset.map(
        tt(tokenizer),
        batched=True,
        remove_columns=["texts"],
        desc="Tokenizing",

    )

    dataset.set_format("torch")

    return dataset
